@@ -1,4 +1,4 @@
-#define DEBUG
+/*#define DEBUG*/
 #include <sys/types.h>
 #include <sys/ioctl.h>
 
@@ -21,6 +21,7 @@
 
 extern struct gway gates[NGATE], *gateway, *gateNGATE;
 extern int ngate;
+extern struct mbuf* mbuf_free;
 
 struct nfile {
 	int n_flags;
@@ -170,8 +171,11 @@ int reent;
 	}
 	we->n_tcb = tp;
 
-	if (!tp->syn_acked)		/* block if not connected */
-		return 1;
+#ifdef DEBUG
+	printf (
+	   "tcp_read: slot = %d, n_len = %d(%d), reent = %d, buffer = 0x%x\n",
+		we->n_slot, we->n_count, we->n_len, reent, we->n_buffer);
+#endif
 
 	if (!reent) {
 		we->n_datawait = FALSE;
@@ -179,6 +183,9 @@ int reent;
 		we->n_count = we->n_len;
 	}
 
+	if (!tp->syn_acked)		/* block if not connected */
+		return 2;
+
 	if (we->n_datawait) {
 	   if (up->uc_rbuf == NULL) {	/* no data available */
 	      if (up->uc_state == 0) {
@@ -187,15 +194,11 @@ int reent;
 		    return 0;
 		 }
 	      }
-	      return 1;
+	      return 3;
 	   }
 	   we->n_datawait = FALSE;
 	}
 
-#ifdef DEBUG
-	printf ("tcp_read: n_slot = %d, n_len = %d(%d), buffer = 0x%x\n",
-		we->n_slot, we->n_count, we->n_len, we->n_buffer);
-#endif
 	while (we->n_len > 0 && !we->n_eol) {
 
 	        if (up->uc_rbuf != NULL) {	/* data available */
@@ -208,7 +211,7 @@ int reent;
 #ifdef DEBUG
 				printf ("   ULOCKed\n");
 #endif
-                        	return 1;
+                        	return 4;
 			}
                         up->uc_flags |= ULOCK;
 
@@ -242,6 +245,9 @@ int reent;
 
                         	if (len == m->m_len) {
         			        we->n_eol = (int)m->m_act;
+#ifdef DEBUG
+					printf ("   free mbuf = 0x%x\n", m);
+#endif
                         		m = m_free(m);
                         		--up->uc_rsize;
                         	} else {
@@ -277,7 +283,7 @@ int reent;
 #endif
         		w_alloc(IURECV, 0, tp, NULL, we);
 			we->n_datawait = TRUE;
-			return 1;
+			return 5;
 		} else {
 			/*
 			 * Net status to return.
@@ -311,17 +317,18 @@ int reent;
 	}
 	we->n_tcb = tp;
 
-	if (!tp->syn_acked)		/* block if not connected */
-		return 1;
+#ifdef DEBUG
+	printf ("tcp_write: slot = %d, n_len = %d(%d), reent = %d\n",
+		we->n_slot, we->n_count, we->n_len, reent);
+#endif
 
 	if (!reent) {
 		we->n_datawait = FALSE;
 		we->n_count = we->n_len;
 	}
-#ifdef DEBUG
-	printf ("tcp_write: n_slot = %d, n_len = %d(%d)\n",
-		we->n_slot, we->n_count, we->n_len);
-#endif
+
+	if (!tp->syn_acked)		/* block if not connected */
+		return 2;
 
 	while (we->n_len > 0 && up->uc_state == 0) {
 		register struct mbuf *n;
@@ -357,13 +364,19 @@ int reent;
 			we->n_datawait = FALSE;
         		if ((m = m_get(1)) == NULL) {
 				/* send what we have, make work req */
-				if (nn.top)
+				if (nn.top) {
         				w_alloc(IUSEND, 0, tp, nn.top, we);
 #ifdef DEBUG
-				printf ("   w_alloc-1: len = %d\n", we->n_len);
+					printf (
+					  "   w_alloc-1: len = %d, bufs = %d\n",
+						we->n_len, bufs + 1);
 #endif
-				return 1;
+				}
+				return 3;
 			}
+#ifdef DEBUG
+			printf ("   get mbuf = 0x%x\n", m);
+#endif
         		m->m_off = MHEAD;
 
 			/* copy user data in */
@@ -392,19 +405,17 @@ int reent;
         		n = m;
         	}
 
-		if (!we->n_datawait) {
-		   w_alloc(IUSEND, 0, tp, nn.top, we);         /* make work req */
+		w_alloc(IUSEND, 0, tp, nn.top, we);         /* make work req */
 #ifdef DEBUG
-		   printf ("   w_alloc-2: len = %d, bufs = %d\n",
-				we->n_len, bufs + 1);
+		printf ("   w_alloc-2: len = %d, bufs = %d\n",
+		        we->n_len, bufs + 1);
 #endif
-		}
 
 		/* if more to send, sleep on more buffers */
 
 		if (we->n_len > 0) {
 		        we->n_datawait = TRUE;
-			return 1;
+			return 4;
 		}
 	}
 	we->n_retval = we->n_count;
@@ -631,12 +642,12 @@ struct netrequest *we;
 	net_t net;
 
 	retval = 0;
-	/* get user connection structure parm */
 
 #ifdef DEBUGOPEN
-	printf ("netopen: n_slot = %d, n_usrfd = %d\n",
+	printf ("netopen: slot = %d, usrfd = %d\n",
 		we->n_slot, we->n_usrfd);
 #endif
+	/* get user connection structure parm */
 	sysnet (SYSCOPYIN, (caddr_t)&const, we->n_buffer,
 		we->n_slot, sizeof(const));
 	cp = (struct con *)&const;
@@ -817,7 +828,7 @@ struct netrequest *we;
 	register struct ucb *up;
 
 #ifdef DEBUGCLOSE
-	printf ("netclose: n_slot = %d, n_usrfd = %d\n",
+	printf ("netclose: slot = %d, usrfd = %d\n",
 		we->n_slot, we->n_usrfd);
 #endif
 	up = nfiles[we->n_slot][we->n_usrfd].n_ucb;
@@ -837,8 +848,7 @@ struct netrequest *we;
 	int retval;
 
 #ifdef DEBUG
-	printf ("netread: n_slot = %d, n_usrfd = %d\n",
-		we->n_slot, we->n_usrfd);
+	printf ("\nnetread: slot = %d, usrfd = %d\n", we->n_slot, we->n_usrfd);
 #endif
 	retval = 0;
 	up = nfiles[we->n_slot][we->n_usrfd].n_ucb;
@@ -848,6 +858,9 @@ struct netrequest *we;
 
 	case UTCP:
 		retval = tcp_read(we, up, 0);
+#ifdef DEBUG
+		printf ("tcp_read: slot = %d, r = %d\n", we->n_slot, retval);
+#endif
 		break;
 
 #ifdef SUPPORTRAW
@@ -874,8 +887,7 @@ struct netrequest *we;
 	int retval;
 
 #ifdef DEBUG
-	printf ("netwrite: n_slot = %d, n_usrfd = %d\n",
-		we->n_slot, we->n_usrfd);
+	printf ("\nnetwrite: slot = %d, usrfd = %d\n", we->n_slot, we->n_usrfd);
 #endif
 	retval = 0;
 	up = nfiles[we->n_slot][we->n_usrfd].n_ucb;
@@ -888,6 +900,10 @@ struct netrequest *we;
 
 		case UTCP:
 			retval = tcp_write(we, up, 0);
+#ifdef DEBUG
+			printf ("tcp_write: slot = %d, r = %d\n",
+				we->n_slot, retval);
+#endif
 			break;
 #ifdef SUPPORTRAW
 		case UIP:
@@ -1247,6 +1263,8 @@ netioctl(we)
 #endif
 	case NETALLSTATS:
 		{
+		   int c;
+      		   struct mbuf *m;
 		   struct getallstats allstats;
 
 		   cnt = 0;
@@ -1317,10 +1335,19 @@ netioctl(we)
 		   }
 		   allstats.ifstats.count = cnt;
 
-		   allstats.ifmbufs.pages = netcb.n_pages;
 		   allstats.ifmbufs.bufs = netcb.n_bufs;
-		   allstats.ifmbufs.hiwat = netcb.n_hiwat;
 		   allstats.ifmbufs.lowat = netcb.n_lowat;
+		   allstats.ifmbufs.hiwat = netcb.n_hiwat;
+
+		   m = mbuf_free;
+		   c = 0;
+		   while (m)
+		   {
+		      c++;
+		      m = m->m_next;
+		   }
+		   allstats.ifmbufs.free = c;
+		   allstats.ifmbufs.max = NMBUF;
 
 		   allstats.nstats.m_drops = netstat.m_drops;
 		   allstats.nstats.ip_badsum = netstat.ip_badsum;
@@ -1354,17 +1381,23 @@ usr_reply()
 	register struct netrequest *we;
 	register struct ucb *up;
 	register struct tcb *tp;
-	int i;
+	int i, r, retval, sendreply;
+
+	retval = 0;
 
 	for (i = 0; i < CMAPSIZ; i++) if (requests[i].n_next) {
 		we = &requests[i];
 		up = we->n_ucb;
 		tp = we->n_tcb;
+		sendreply = 1;
+
 		switch (we->n_request) {
 
 		case SYSOPEN:
-			if (up->uc_state == 0 && !tp->syn_acked)
-          			return 1;
+			if (up->uc_state == 0 && !tp->syn_acked) {
+				sendreply = 0;
+          			break;
+			}
 
 			if (up->uc_state != 0) {
 				if (up->uc_state & UINTIMO)
@@ -1381,22 +1414,37 @@ usr_reply()
 				we->n_error = EINTR;
 			if (we->n_error == 0)
 				we->n_retval = we->n_usrfd;
+			retval = 1;
 			break;
 
 		case SYSREAD:
-			if (tcp_read (we, up, 1))
-				return 1;
+			if ((r = tcp_read (we, up, 1)) != 0) {
+				we->n_next = (struct netrequest *)r;
+				sendreply = 0;
+			}
+#ifdef DEBUG
+			printf ("tcp_read: slot = %d, r = %d\n",
+				we->n_slot, r);
+#endif
+			retval = 1;
 			break;
 
 		case SYSWRITE:
-			if (tcp_write (we, up, 1))
-				return 1;
+			if ((r = tcp_write (we, up, 1)) != 0) {
+				sendreply = 0;
+				we->n_next = (struct netrequest *)r;
+			}
+#ifdef DEBUG
+			printf ("tcp_write: slot = %d, r = %d\n",
+				we->n_slot, r);
+#endif
+			retval = 1;
 			break;
 
 		case SYSCLOSE:
 #if 0
 			if (tcp_close (we, up, 1))
-				return 1;
+				sendreply = 0;
 #endif
 			break;
 
@@ -1404,31 +1452,35 @@ usr_reply()
 		default: ;
 		}
 
-		we->n_next = 0; /* resusing field for request in-process */
- 		if (we->n_error) {
-#ifdef DEBUGREQ
-	   		printf ("usr_reply: pid = %d, n_error = %d\n",
-				we->n_pid, we->n_error);
+		if (sendreply) {
+			/* resusing field for request in-process */
+			we->n_next = 0;
+	 		if (we->n_error) {
+		   		we->n_retval = -1;
+			}
+#ifdef DEBUG
+		        printf ("Reply: slot = %d, error = %d, retval = %d\n",
+				we->n_slot, we->n_error, we->n_retval);
 #endif
-		   	we->n_retval = -1;
+			sysnet (SYSREPLY, we->n_pid, we->n_proc,
+				we->n_error, we->n_retval);
+			retval = 1;
 		}
-		sysnet (SYSREPLY, we->n_pid, we->n_proc,
-			we->n_error, we->n_retval);
 	}
-	return 0;
+	return retval;
 }
 
 int
 usr_request (we)
 struct netrequest *we;
 {
+	memcpy (&requests[we->n_slot], we, sizeof (struct netrequest));
+	we = &requests[we->n_slot];
+
 	we->n_error = 0;
 	we->n_retval = 0;
 	we->n_next = 0;
 
-	memcpy (&requests[we->n_slot], we, sizeof (struct netrequest));
-	we = &requests[we->n_slot];
-
 	switch (we->n_request) {
 	case SYSOPEN:
 		if (netopen (we)) {
@@ -1464,11 +1516,12 @@ struct netrequest *we;
 	}
 
 	if (we->n_error) {
-#ifdef DEBUGREQ
-	   printf ("usr_request: n_error = %d\n", we->n_error);
-#endif
 	   we->n_retval = -1;
 	}
+#ifdef DEBUG
+        printf ("Reply: slot = %d, error = %d, retval = %d\n",
+		we->n_slot, we->n_error, we->n_retval);
+#endif
 	sysnet (SYSREPLY, we->n_pid, we->n_proc, we->n_error, we->n_retval);
 	return 0;
 }
