@@ -58,11 +58,16 @@ loop:
 		write(1, t, 1);
 	}
 	c = *fmt++;
-	if(c == 'x')
+	if(c == 'x') {
 		printx(*adx);
-	else if(c == 'd' || c == 'l' || c == 'o')
+	} else if(c == 'd' || c == 'l' || c == 'o') {
+		if (*adx < 0) {
+		   t[0] = '-';
+		   write (1, t, 1);
+		   *adx = - *adx;
+		}
 		printn(*adx, c=='o'? 8: 10);
-	else if(c == 's') {
+	} else if(c == 's') {
 		s = (char*) *adx;
 		while(c = *s++) {
 			t[0] = c;
